[Setting name="FirstRun" hidden]
bool S_FirstRun = true;

[Setting name="Display skillpoints while playing"]
bool S_DisplayWhilePlaying = false;

[Setting name="Automatically reload skillpoints on game start"]
bool S_ReloadOnStart = true;

[Setting name="Text colour" color]
vec3 S_TextColour = vec3(0,1,0);

CachedSkillpoints cache;

int fontOswald = nvg::LoadFont("fonts/Oswald-Regular.ttf");
int fontOswaldItalic = nvg::LoadFont("fonts/Oswald-Italic.ttf");
nvg::Texture@ spIconImage = nvg::LoadTexture("textures/sp_icon.png");
nvg::Paint spIcon = nvg::TexturePattern(vec2(0, 0), vec2(128, 128), 0, spIconImage, 1);
int originalSkillPoints;

int envi = 1;
int series = 1;

uint page = 0;

vec4 colWhite = Text::ParseHexColor("#e0e0e0");
vec4 colGold = Text::ParseHexColor("#e5d800");

// update sp value every second to prevent lag
string prevTime = "";

bool window_visible = false;
bool firstrun_visible = false;

void Main() {
    cache.Load();
    yield();
    auto app = cast<CTrackMania>(GetApp());
    if (S_ReloadOnStart && !S_FirstRun) {
        //wait for log in
        while (app.ManiaPlanetScriptAPI.MasterServer_MSUsers.Length == 0) {
            yield();
        }
        startnew(ReloadAllMapsSkillpointsAsync);
    }
    while (true) {
        yield();
        if (prevTime != app.OSUTCTime) {
            cache.Save();
            yield();
            prevTime = app.OSUTCTime;
        } 
    }
}


void RenderMenu() {
    auto app = cast<CTrackMania>(GetApp());
    auto menu = cast<CTrackManiaMenus>(app.MenuManager);
    auto ManiaApp = menu.MenuCustom_CurrentManiaApp;

    if (UI::BeginMenu("Turbo Skillpoints")) {
        if (UI::MenuItem("Reload all maps", "", false)) {
            startnew(ReloadAllMapsSkillpointsAsync);
        }

        if (UI::MenuItem("Top maps", "", window_visible)) {
            window_visible = !window_visible;
        }

        UI::EndMenu();
    }
}
void RenderInterface() {
    if (!window_visible) {
        return;
    }
    auto app = cast<CTrackMania>(GetApp());
    auto menu = cast<CTrackManiaMenus>(app.MenuManager);
    auto ManiaApp = menu.MenuCustom_CurrentManiaApp;

    if (window_visible) {
        auto uidJson = Json::FromFile("data/mapUids.json");
        auto campaignMapIds = uidJson.Get("uids", Json::Array());

        MapInstance@[] sortedMaps;

        if (cache.list.Length > 0) {
            for (uint i = 0; i < cache.list.Length; i++) {
                sortedMaps.InsertLast(cache.list[i]);
            }
        }
        
        uint pageStartIndex = 10*page;
        uint pageEndIndex = pageStartIndex+10;
        uint maxPage = Math::Floor(sortedMaps.Length/10.0)-1;
        page = Math::Clamp(page, 0, maxPage);
        sortedMaps.SortDesc();
        UI::SetNextWindowSize(165, 295);
        if (UI::Begin("Top maps", window_visible, UI::WindowFlags::NoCollapse | UI::WindowFlags::NoResize)) {
            // UI::Text("Top 10 maps:");
            if (UI::BeginChild("list-container", vec2(150*UI::GetScale(),225*UI::GetScale()))) {
                if (UI::BeginTable("topPointsTable", 3, UI::TableFlags::RowBg)) {
                    UI::TableSetupColumn("#", UI::TableColumnFlags::WidthFixed, 25*UI::GetScale());
                    UI::TableSetupColumn("Map", UI::TableColumnFlags::WidthFixed, 35*UI::GetScale());
                    UI::TableSetupColumn("SP", UI::TableColumnFlags::WidthFixed, 70*UI::GetScale());
                    UI::TableHeadersRow();
                    for (uint i = pageStartIndex; i < Math::Min(pageEndIndex, sortedMaps.Length); i++) {
                        string placenumber = PadNumber(i+1, 0);
                        uint mapIndex = 0;
                        for (uint j = 0; j < campaignMapIds.Length; j++) {
                            if (campaignMapIds[j] == sortedMaps[i].uid) {
                                mapIndex = j;
                                break;
                            }
                        }
    
                        string mapnumber = PadNumber(mapIndex+1, 3);
                        
                        UI::TableNextRow();
                        UI::TableNextColumn();
                        UI::Text(""+(placenumber));
                        UI::TableNextColumn();
                        UI::Text("#"+Text::FormatOpenplanetColor(getMapColor(mapIndex))+mapnumber);
                        UI::TableNextColumn();
                        UI::Text(Text::FormatOpenplanetColor(S_TextColour)+sortedMaps[i].value);
/*                         UI::Text(placenumber+
                                 ". #"+
                                 Text::FormatOpenplanetColor(getMapColor(mapIndex))+
                                 mapnumber+
                                 "\\$g: "+
                                 Text::FormatOpenplanetColor(S_TextColour)+sortedMaps[i].value+
                                 " SP"); */
                    }
                    UI::EndTable();
                }

                UI::EndChild();
                UI::BeginDisabled(page == 0);
                if (UI::Button(Icons::Backward)) {
                    page = 0;
                }
                UI::SameLine();
                if (UI::Button(Icons::ChevronLeft)) {
                    page--;
                }
                UI::EndDisabled();
                UI::BeginDisabled(page == maxPage);
                UI::SameLine();
                if (UI::Button(Icons::ChevronRight)) {
                    page++;
                }
                UI::SameLine();
                if (UI::Button(Icons::Forward)) {
                    page = 19;
                }
                UI::EndDisabled();
            }
        }
    }
    UI::End();


}

void Render() {
    auto app = cast<CTrackMania>(GetApp());
    auto menu = cast<CTrackManiaMenus>(app.MenuManager);
    auto playground = cast<CTrackManiaRaceNew>(app.CurrentPlayground);
    auto ManiaApp = menu.MenuCustom_CurrentManiaApp;

    if (app.ManiaPlanetScriptAPI.MasterServer_MSUsers.Length == 0) return;

    if (S_FirstRun) {
        UI::SetNextWindowSize(295, 140);
        if (UI::Begin("Turbo Skillpoints##FirstRun", UI::WindowFlags::NoCollapse | UI::WindowFlags::NoResize)) {
            UI::TextWrapped("You are running the plugin for the first time. Would you like to load skillpoints on all campaign maps? You will be able to load all maps again in the plugins menu.");
            if (UI::Button("Yes")) {
                S_FirstRun = false;
                startnew(ReloadAllMapsSkillpointsAsync);
            }
            UI::SameLine();
            if (UI::Button("No")) {
                S_FirstRun = false;
            }
        }
        UI::End();
    }

    uint mapRank = 0;
    uint mapTotal = 1;
    // in map
    if (app.Challenge !is null && playground !is null) {     //  Regular Campaign                                VR Campaign  
        bool campaignMap = app.Challenge.MapName != "0" && (app.Challenge.MapInfo.AuthorLogin == "Nadeo" || app.Challenge.MapInfo.AuthorLogin == "nadeolabs");
        if (campaignMap) {
        
            mapRank = ManiaApp.ScoreMgr.MapLeaderBoard_GetPlayerRanking(
                    app.ManiaPlanetScriptAPI.MasterServer_MSUsers[0].Id, 
                    app.Challenge.EdChallengeId, 
                    "World");
            mapTotal = ManiaApp.ScoreMgr.MapLeaderBoard_GetPlayerCount(app.Challenge.EdChallengeId, "World");

            string spText = ""+getSkillPoints(mapRank, mapTotal);
            if (spText == "-1") spText = "Loading...";
            
            if (playground.Players.Length > 0) {
                auto player = cast<CTrackManiaPlayer>(playground.Players[0]);
                if (player.RaceState !=  CTrackManiaPlayer::ERaceState::Finished) {
                    if (getSkillPoints(mapRank, mapTotal) != -1) {
                        originalSkillPoints = getSkillPoints(mapRank, mapTotal);
                        if (S_DisplayWhilePlaying) {
                            nvg::FontSize(30.0);
                            nvg::FontFace(fontOswald);
                            drawImage(vec2(12, 0), vec2(60, 60), spIconImage);
                            drawTextWithShadow(82, 40, vec4(S_TextColour,1), spText);

                        }
                    } else {
                        originalSkillPoints = 0;
                    }
                } else if (player.RaceState ==  CTrackManiaPlayer::ERaceState::Finished && playground.UIConfigs[0].UISequence != CGamePlaygroundUIConfig::EUISequence::EndRound)  {
                    int spGain = getSkillPoints(mapRank,mapTotal) - originalSkillPoints;
                    spText = ""+getSkillPoints(mapRank, mapTotal)+" SP";
                    nvg::FontSize(35.0);
                    vec2 bounds = nvg::TextBounds(spText);
                    if (spGain > 0) {
                        nvg::FontFace(fontOswaldItalic);
                        drawTextWithShadow((Draw::GetWidth()/2) - bounds.x/2, Draw::GetHeight()/1.655, colWhite, spText);
                        bounds = nvg::TextBounds("+"+spGain);
                        drawTextWithShadow((Draw::GetWidth()/2) - bounds.x/2, Draw::GetHeight()/1.565, colGold, "+"+spGain);

                    } /* else {
                        drawTextWithShadow((Draw::GetWidth()/2) - bounds.x/2, Draw::GetHeight()/1.610, colWhite, spText);
                    } */
                }
            }


        }


    } else /* in menu */ {
        vec2[] menuImagePositions = {
            vec2(Draw::GetWidth()/7.875, Draw::GetHeight()/1.766),
            vec2(Draw::GetWidth()/3.585, Draw::GetHeight()/1.766),
            vec2(Draw::GetWidth()/2.335, Draw::GetHeight()/1.766),
            vec2(Draw::GetWidth()/1.730, Draw::GetHeight()/1.766),
            vec2(Draw::GetWidth()/1.370, Draw::GetHeight()/1.766),
            vec2(Draw::GetWidth()/7.875, Draw::GetHeight()/1.226),
            vec2(Draw::GetWidth()/3.585, Draw::GetHeight()/1.226),
            vec2(Draw::GetWidth()/2.335, Draw::GetHeight()/1.226),
            vec2(Draw::GetWidth()/1.730, Draw::GetHeight()/1.226),
            vec2(Draw::GetWidth()/1.370, Draw::GetHeight()/1.226),
        };
        vec2[] menuTextPositions = {
            vec2(Draw::GetWidth()/6.575, Draw::GetHeight()/1.680),
            vec2(Draw::GetWidth()/3.285, Draw::GetHeight()/1.680),
            vec2(Draw::GetWidth()/2.200, Draw::GetHeight()/1.680),
            vec2(Draw::GetWidth()/1.655, Draw::GetHeight()/1.680),
            vec2(Draw::GetWidth()/1.325, Draw::GetHeight()/1.680),
            vec2(Draw::GetWidth()/6.575, Draw::GetHeight()/1.185),
            vec2(Draw::GetWidth()/3.285, Draw::GetHeight()/1.185),
            vec2(Draw::GetWidth()/2.200, Draw::GetHeight()/1.185),
            vec2(Draw::GetWidth()/1.655, Draw::GetHeight()/1.185),
            vec2(Draw::GetWidth()/1.325, Draw::GetHeight()/1.185),
        };

        auto uidJson = Json::FromFile("data/mapUids.json");

        auto campaignMapIds = uidJson.Get("uids", Json::Array());

        // UILayer 11 is campaign map selection
        if (ManiaApp.UILayers.Length < 17) return;
        auto Layer = cast<CGameUILayer>(ManiaApp.UILayers[11]);
        // UILayer 17 is campaign background (show total when it's visible)
        auto LayerBackground = cast<CGameUILayer>(ManiaApp.UILayers[17]);
        auto LayerFrame = cast<CGameManialinkFrame>(Layer.LocalPage.MainFrame.Controls[0]);
        auto FrameContentContainer = cast<CGameManialinkFrame>(LayerFrame.Controls[4]);
        auto FrameContent = cast<CGameManialinkFrame>(FrameContentContainer.Controls[1]);
        auto FrameButtonsContainer = cast<CGameManialinkFrame>(FrameContent.Controls[2]);
        auto FrameButtons = cast<CGameManialinkFrame>(FrameButtonsContainer.Controls[1]);
        
        auto LabelSeries = cast<CGameManialinkLabel>(FrameButtons.Controls[3]);
        
        auto FrameCanyon = cast<CGameManialinkFrame>(FrameButtons.Controls[20]);
        auto FrameValley = cast<CGameManialinkFrame>(FrameButtons.Controls[25]);
        auto FrameLagoon = cast<CGameManialinkFrame>(FrameButtons.Controls[30]);
        auto FrameStadium = cast<CGameManialinkFrame>(FrameButtons.Controls[35]);

        auto FrameBackground = cast<CGameManialinkFrame>(LayerBackground.LocalPage.MainFrame.Controls[0]);
        auto FrameBackgroundContent = cast<CGameManialinkFrame>(FrameBackground.Controls[0]);
        auto FrameSuperSolo = cast<CGameManialinkFrame>(FrameBackgroundContent.Controls[1]);

        auto QuadSuperSoloSky = cast<CGameManialinkQuad>(FrameSuperSolo.Controls[0]);


        uint mapStartPos = 0;

        if (LabelSeries.Value.Contains("White")) {
            series = 1;
        } else if (LabelSeries.Value.Contains("Green")) {
            series = 2;
        } else if (LabelSeries.Value.Contains("Blue")) {
            series = 3;
        } else if (LabelSeries.Value.Contains("Red")) {
            series = 4;
        } else if (LabelSeries.Value.Contains("Black")) {
            series = 5;
        }

        if (Math::Round(FrameCanyon.AbsolutePosition_V3.x) == -120) {
            envi = 1;
        } else if (Math::Round(FrameValley.AbsolutePosition_V3.x) == -120) {
            envi = 2;
        } else if (Math::Round(FrameLagoon.AbsolutePosition_V3.x) == -120) {
            envi = 3;
        } else if (Math::Round(FrameStadium.AbsolutePosition_V3.x) == -120) {
            envi = 4;
        }

        mapStartPos = ((envi-1)*10)+((series-1)*40);
        
        nvg::FontFace(fontOswald);

        if (Layer.IsVisible) {
            for (uint i = 0; i < menuTextPositions.Length-1; i++ ) {
                string curMap = campaignMapIds[i+mapStartPos];

                mapRank = ManiaApp.ScoreMgr.MapLeaderBoard_GetPlayerRanking(
                app.ManiaPlanetScriptAPI.MasterServer_MSUsers[0].Id, 
                curMap, 
                "World");
                mapTotal = ManiaApp.ScoreMgr.MapLeaderBoard_GetPlayerCount(curMap, "World");
                string spText = ""+getSkillPoints(mapRank, mapTotal);

                // this part is laggy
                if (getSkillPoints(mapRank, mapTotal) == -1) {
                    if (cache.Exists(curMap) ) {
                        spText = "" + cache.GetPoints(curMap);
                    }
                } else {
                    if (cache.Exists(curMap)) {
                        if (cache.GetPoints(curMap) != getSkillPoints(mapRank,mapTotal)) cache.UpdateValue(curMap, getSkillPoints(mapRank,mapTotal));
                    } else {
                        cache.Add(curMap, getSkillPoints(mapRank, mapTotal));
                    }
                    spText = ""+getSkillPoints(mapRank, mapTotal);
                    // 
                }

                if (spText == "-1") spText = "Loading...";

                nvg::FontSize(25.0);
                drawImage(menuImagePositions[i], vec2(45, 45), spIconImage);
                drawTextWithShadow(menuTextPositions[i], vec4(S_TextColour,1), spText);
            }

        }

        if (LayerBackground.IsVisible) {
            nvg::FontSize(30.0);
            nvg::FontFace(fontOswald);
            int totalSP = cache.GetTotalPoints();
            if (!QuadSuperSoloSky.Visible) {
                drawImage(vec2(Draw::GetWidth()/10, Draw::GetHeight()/31), vec2(60, 60), spIconImage);
                drawTextWithShadow(vec2(Draw::GetWidth()/7.5, Draw::GetHeight()/14.5), vec4(S_TextColour,1), ""+totalSP);
            } else { // super solo has different background image
                drawImage(vec2(Draw::GetWidth()/9.6, Draw::GetHeight()/52), vec2(60, 60), spIconImage);
                drawTextWithShadow(vec2(Draw::GetWidth()/7.2, Draw::GetHeight()/17.5), vec4(S_TextColour,1), ""+totalSP);
            }

        }


    }
}

void ReloadAllMapsSkillpointsAsync() {
    auto app = cast<CTrackMania>(GetApp());
    auto menu = cast<CTrackManiaMenus>(app.MenuManager);
    auto ManiaApp = menu.MenuCustom_CurrentManiaApp;

    auto uidJson = Json::FromFile("data/mapUids.json");
    auto campaignMapIds = uidJson.Get("uids", Json::Array());

    UI::ShowNotification("Turbo Skillpoints", "Loading skillpoints on all maps, this may take a while...", 10000);

    for (uint i = 0; i < campaignMapIds.Length; i++) {
        string curMap = campaignMapIds[i];
        int mapRank = ManiaApp.ScoreMgr.MapLeaderBoard_GetPlayerRanking(
        app.ManiaPlanetScriptAPI.MasterServer_MSUsers[0].Id, 
        curMap, 
        "World");
        int mapTotal = ManiaApp.ScoreMgr.MapLeaderBoard_GetPlayerCount(curMap, "World");

        // wait max 5000 seconds per map
        int timeout = 0;

        while (getSkillPoints(mapRank,mapTotal) == -1 && timeout < 5000) {
            yield();
            // retry
            mapRank = ManiaApp.ScoreMgr.MapLeaderBoard_GetPlayerRanking(
            app.ManiaPlanetScriptAPI.MasterServer_MSUsers[0].Id, 
            curMap, 
            "World");
            mapTotal = ManiaApp.ScoreMgr.MapLeaderBoard_GetPlayerCount(curMap, "World");
            if (prevTime != app.OSUTCTime) {
                timeout += 1000;
            }
        }

        if (timeout >= 5000) {
            // because it can happen if there's no pb on the map, don't display a message           
            // UI::ShowNotification("Turbo Skillpoints", "Failed to load SP on map #"+PadNumber(i+1, 3), vec4(1,0,0,1) ,2000);
            continue;
        }

        if (cache.Exists(curMap)) {
            if (cache.GetPoints(curMap) != getSkillPoints(mapRank,mapTotal)) cache.UpdateValue(curMap, getSkillPoints(mapRank,mapTotal));
        } else {
            cache.Add(curMap, getSkillPoints(mapRank, mapTotal));
        }
        // UI::ShowNotification("Turbo Skillpoints", "Loaded SP on map #"+PadNumber(i+1, 3), vec4(0,0.4,0,1) ,2000);
    }
    UI::ShowNotification("Turbo Skillpoints", "Done loading SP on all maps", vec4(0,0.7,0,1) ,5000);
    yield();
}

int getSkillPoints(int rank, int total) {
    if (rank==0) return -1;
    return ((total-rank)*100)/rank;
}

void drawTextWithShadow(float x, float y, vec4 color, const string &in text) {
    nvg::FillColor(vec4(0.0, 0.0, 0.0, 1.0));
    nvg::Text(x+2, y+2, "" + text);
    nvg::FillColor(color);
    nvg::Text(x, y, "" + text);
}

void drawTextWithShadow(vec2 pos, vec4 color, const string &in text) {
    nvg::FillColor(vec4(0.0, 0.0, 0.0, 1.0));
    nvg::Text(pos.x+2, pos.y+2, "" + text);
    nvg::FillColor(color);
    nvg::Text(pos, "" + text);
}

void drawImage(vec2 origin, vec2 size, nvg::Texture @texture) {
     nvg::BeginPath();
    nvg::Rect(origin, size);
    nvg::FillPaint(nvg::TexturePattern(origin, size, 0, texture, 1));
    nvg::Fill();
    nvg::ClosePath();
}

void drawRect(vec2 origin, vec2 size, vec4 color) {
     nvg::BeginPath();
    nvg::Rect(origin, size);
    nvg::FillColor(color);
    nvg::Fill();
    nvg::ClosePath();
}

string PadNumber(int number, uint places) {
    string numberText = ""+number;
    string text = "";

    int toPad = Math::Max(places-numberText.Length, 0);

    for (uint i = 0; i < toPad; i++) {
        text += "0";
    }
    text+=""+number;
    return text;
} 

vec3 getMapColor(int number) {
    vec4 color = vec4(0,0,0,1);
    if (number <= 40) {
        color = Text::ParseHexColor("#ffffff");
    } else if (number <= 80) {
        color = Text::ParseHexColor("#44aa22");
    } else if (number <= 120) {
        color = Text::ParseHexColor("#2266cc");
    } else if (number <= 160) {
        color = Text::ParseHexColor("#ff3311");
    } else if (number <= 200) {
        color = Text::ParseHexColor("#777777");
    } 
    return vec3(color.xyz);
}